<html>
<head>
    <title>Authentication</title>
</head>
    <body>
        <div>
            <p>
                A client is trying to {{ description | e }}. To confirm this action,
                <a href="{{ redirect_url | e }}">re-authenticate with single sign-on</a>.
                If you did not expect this, your account may be compromised!
            </p>
        </div>
    </body>
</html>
