html {
    height: 100%;
}

body {
    height: 100%;
    font-family: "Myriad Pro", "Myriad", Helvetica, Arial, sans-serif;
    font-size: 12pt;
    margin: 0px;
}

h1 {
    font-size: 20pt;
}

a:link    { color: #666; }
a:visited { color: #666; }
a:hover   { color: #000; }
a:active  { color: #000; }

input {
   margin: 5px;
}

textbox, input[type="text"], input[type="password"] {
   width: 90%;
}

form {
    text-align: center;
    margin: 10px 0 0 0;
}

.login_flow {
    width: 300px;
    text-align: left;
    padding: 10px;
    margin-bottom: 40px;

    -webkit-border-radius: 10px;
    -moz-border-radius: 10px;
    border-radius: 10px;

    -webkit-box-shadow: 0px 0px 20px 0px rgba(0,0,0,0.15);
    -moz-box-shadow: 0px 0px 20px 0px rgba(0,0,0,0.15);
    box-shadow: 0px 0px 20px 0px rgba(0,0,0,0.15);

    background-color: #f8f8f8;
    border: 1px #ccc solid;
}
