<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>SSO redirect confirmation</title>
</head>
    <body>
        <p>The application at <span style="font-weight:bold">{{ display_url | e }}</span> is requesting full access to your <span style="font-weight:bold">{{ server_name }}</span> Matrix account.</p>
        <p>If you don't recognise this address, you should ignore this and close this tab.</p>
        <p>
            <a href="{{ redirect_url | e }}">I trust this address</a>
        </p>
    </body>
</html>